// This file is part of mzCalc.
//
// This program is free software; you can redistribute it and/or modify it under
// the terms of the GNU General Public License version 2, as published by the
// Free Software Foundation.
//
// This program is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
// details.
//
// You should have received a copy of the GNU General Public License along with
// mzCalc. If not, see <http://www.gnu.org/licenses/>.
//
// This work is Copyright (C) 2016 by Josse van der Plaat
//
// Individual portions may be copyright by individual contributors, and are
// included in this collective work with permission of the copyright owners.

#include "fleetcomposition.h"

#include "stats.h"
#include "shiptype.h"

#include "logger.h"

#include <algorithm>

std::array<double, (Class::Battleship + 1 - Class::Fighter)> targetAllocation(Class clazz, Verbosity verbosity) {
    std::array<double, (Class::Battleship + 1 - Class::Fighter)> targets;
    // This is the default allocation assuming we can hit all targets. These are
    // ratios.
    double sum = 0.0;
    switch (clazz.clazz) {
    case Class::Fighter:
    case Class::Corvette: {
        targets = { 4.0, 4.0, 2.0, 2.0, 1.0, 1.0 };
        sum = 14.0;
    } break;
    case Class::Frigate:
    case Class::Destroyer: {
        targets = { 3.0, 3.0, 2.0, 2.0, 1.0, 1.0 };
        sum = 12.0;
    } break;
    case Class::Cruiser:
    case Class::Battleship: {
        targets = { 1.0, 1.0, 2.0, 2.0, 1.0, 1.0 };
        sum = 8.0;
    } break;
    default: {
        Logger(verbosity).error("Pod in invalid class %s found!\n",
                                clazz.toString().substr(0,2).c_str());
    }
    }
    std::transform(targets.begin(), targets.end(),
                   targets.begin(),
                   [sum](double a) -> double {
                       return a / sum;
                   });
    return targets;
}

std::array<double, (Class::Battleship + 1 - Class::Fighter)> targetAllocation(const Fleet& fleet) {
    Cost fleetCost = fleet.fleetCost();
    std::array<double, (Class::Battleship + 1 - Class::Fighter)> targets;
    std::fill(targets.begin(), targets.end(), 0.0);
    for (const BattleFormation& bf : fleet.battleFormations) {
        for (std::size_t i = 0; i < MAX_TARGETS; ++i) {
            Class t = bf.shipType.target(i);
            if (t.isShip()) {
                targets[t.clazz] += double(bf.atStartOfBattle * bf.shipType.cost * effs[i]) / 100.0 / fleetCost;
            }
        }
    }
    return targets;
}

// If FROM does not contain at least 1 ShipType that targets CLAZZ, return
// nullptr. Otherwise, return the ShipType in FROM that is 'best' against CLAZZ.
ShipType* bestAnti(Class clazz, std::vector<ShipType*> from) {
    ShipType* best = nullptr;
    for (ShipType* ship : from) {
        std::size_t t = ship->targets(clazz);
        if (t < MAX_TARGETS) {
            // Prefer any ship that targets clazz over nothing. Otherwise,
            // prefer low init over high init. Otherwise, prefer T(n) over T(m :
            // m > n).
            if (!best || ship->init < best->init || t < best->targets(clazz)) {
                best = ship;
            }
        }
    }
    return best;
}

void createAttackFleet(Stats& stats, Fleet& fleet, const ShipType& pod, bool enableResourceTheft, Verbosity verbosity) {
    Logger log(verbosity);
    Cost fleetCost = 100000000;

    NumberOfShips neededToCap = 500 * stats["Asteroid"]->armor / pod.guns;
    fleet.addBattleFormation(pod, neededToCap);
    log.debug("  - Add  %6ld %-8s as pod,      firing at Ro, No costing %4.1fm\n",
              neededToCap,
              pod.name.c_str(),
              double(neededToCap * pod.cost) / 1000000.0);

    if (enableResourceTheft) {
        Maybe<ShipType> resourceStealer = stats.findResourceStealer(pod.race, pod.clazz);
        if (resourceStealer) {
            NumberOfShips count = 2 * neededToCap * pod.cost / resourceStealer->cost;
            fleet.addBattleFormation(*resourceStealer, count);
        }
    }

    std::vector<ShipType*> escorts = stats.findCombatShips(pod.race, pod.clazz);
    auto idealAllocation = targetAllocation(pod.clazz, verbosity);
    std::array<double, (Class::Battleship + 1 - Class::Fighter)> actualAllocation;
    for (Class::C c = Class::Fighter; c <= Class::Battleship; c = Class::C(c + 1)) {
        Cost needed = fleetCost * (idealAllocation[c] - actualAllocation[c]);
        ShipType* s = bestAnti(c, escorts);
        if (needed > 0 && s) {
            NumberOfShips toIdeal = 100 * needed / s->cost / effs[s->targets(c)];
            fleet.addBattleFormation(*s, toIdeal);
            for (std::size_t i = 0; i < MAX_TARGETS; ++i) {
                Class t = s->target(i);
                if (t.isShip()) {
                    actualAllocation[t.clazz] += double(toIdeal * s->cost * effs[i]) / 100.0 / fleetCost;
                }
            }
        }
    }

    // Finally, make sure we haven't gone over or stayed under our
    // budget (that being 2/3 of our full fleet budget).
    std::fill(actualAllocation.begin(), actualAllocation.end(), 0.0);
    Cost actualCost = fleet.fleetCost();
    for (BattleFormation& bf : fleet.battleFormations) {
        bf.atStartOfBattle = bf.atStartOfBattle * 2 * fleetCost / actualCost / 3;
        for (std::size_t i = 0; i < MAX_TARGETS; ++i) {
            Class t = bf.shipType.target(i);
            if (t.isShip()) {
                actualAllocation[t.clazz] += double(bf.atStartOfBattle * bf.shipType.cost * effs[i]) / 100.0 / fleetCost;
            }
        }
        if (bf.shipType.type != GunType::Pod) {
            log.debug("  - With %6ld %-8s as escort,   firing at %s, %s costing %4.1fm\n",
                      bf.atStartOfBattle,
                      bf.shipType.name.substr(0,8).c_str(),
                      bf.shipType.target(0).toString().substr(0,2).c_str(),
                      bf.shipType.target(1).toString().substr(0,2).c_str(),
                      double(bf.atStartOfBattle * bf.shipType.cost) / 1000000.0);
        }
    }
}

void addDefenseShips(Stats& stats, Fleet& fleet, Verbosity verbosity) {
    Logger log(verbosity);

    Cost fleetCost = 100000000;
    Cost actualCost = fleet.fleetCost();
    auto actualAllocation = targetAllocation(fleet);
    int missingTargets = std::count(actualAllocation.begin(), actualAllocation.end(), 0.0);

    for (Class::C c = Class::Fighter; c <= Class::Battleship; c = Class::C(c + 1)) {
        if (actualAllocation[c] == 0.0) {
            auto potentials = stats.findCombatShips(fleet.firstPod()->shipType.race);
            ShipType* s = bestAnti(c, potentials);
            // It is more efficient to spend resources on one ship with two
            // targets than to spend it on two ships with one target each.
            int numberOfTargets = 0;
            for (std::size_t i = 0; i < MAX_TARGETS; ++i) {
                Class t = s->target(i);
                if (t.isShip() && actualAllocation[t.clazz] == 0.0) {
                    numberOfTargets++;
                }
            }
            NumberOfShips needed = (fleetCost - actualCost) * numberOfTargets / missingTargets / s->cost;
            fleet.addBattleFormation(*s, needed);
            for (std::size_t i = 0; i < MAX_TARGETS; ++i) {
                Class t = s->target(i);
                if (t.isShip()) {
                    actualAllocation[t.clazz] += double(needed * s->cost * effs[i]) / 100.0 / fleetCost;
                }
            }
            log.debug("  - With %6ld %-8s for defense, firing at %s, %s costing %4.1fm\n",
                      needed,
                      s->name.substr(0,8).c_str(),
                      s->target(0).toString().substr(0,2).c_str(),
                      s->target(1).toString().substr(0,2).c_str(),
                      double(needed * s->cost) / 1000000.0);
        }
    }
}

std::vector<Fleet> fleetCompositions(Stats& stats, bool enableResourceTheft, Verbosity verbosity) {
    std::vector<Fleet> fleets;
    Logger log(verbosity);
    for (auto& pair : stats) {
        const ShipType& pod = pair.second;
        if (pod.type == GunType::Pod) {
            log.info("* %s\n",
                     pod.name.c_str());

            fleets.emplace_back(pod.name, 0, Side::Defense);
            Fleet& fleet = fleets.back();

            createAttackFleet(stats, fleet, pod, enableResourceTheft, verbosity);
            addDefenseShips(stats, fleet, verbosity);
            fleet.race  = pod.race;
            fleet.value = fleet.fleetCost() / 90;
            fleet.xp    = fleet.value / 120;
            fleet.score = fleet.value + 60 * fleet.xp;

            log.trace("  - Total value = %.1f\n",
                      fleet.fleetCost() / 100 / 1000000.0);

            auto idealAllocation = targetAllocation(pod.clazz, verbosity);
            auto actualAllocation = targetAllocation(fleet);
            log.trace("  - Ideal  allocation for %s: ",
                      pod.clazz.toString().substr(0, 2).c_str());
            for (std::size_t i = 0; i < idealAllocation.size(); ++i) {
                log.trace("%s:%.2f, ",
                          Class(Class::C(i)).toString().substr(0, 2).c_str(),
                          idealAllocation[i]);
            }
            log.trace("\n  - Actual allocation for %s: ",
                      pod.clazz.toString().substr(0, 2).c_str());
            for (std::size_t i = 0; i < actualAllocation.size(); ++i) {
                log.trace("%s:%.2f, ",
                          Class(Class::C(i)).toString().substr(0, 2).c_str(),
                          actualAllocation[i]);
            }
            log.trace("\n");
        }
    }
    return fleets;
}
