// This file is part of mzCalc.
//
// This program is free software; you can redistribute it and/or modify it under
// the terms of the GNU General Public License version 2, as published by the
// Free Software Foundation.
//
// This program is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
// details.
//
// You should have received a copy of the GNU General Public License along with
// mzCalc. If not, see <http://www.gnu.org/licenses/>.
//
// This work is Copyright (C) 2016 by Josse van der Plaat
//
// Individual portions may be copyright by individual contributors, and are
// included in this collective work with permission of the copyright owners.

#include "parser.h"
#include "fleetcomposition.h"
#include "analyzestats.h"
#include "fillbattle.h"
#include "bcalc.h"

#include "battle.h"

#include "logger.h"

#include <getopt.h>
#include <sys/stat.h>
#include <cstring>

#include "version.h"

struct Options {
    bool analyze = false;
    bool calculate = false;
    bool exportCompositions = false;

    Verbosity verbosity = Verbosity::Warning;
    bool help = false;
    bool version = false;

    std::string statsFile = "stats.txt";
    bool repeatSameTargetIndex = true;
    bool enableResourceTheft = false;
    bool import = false;
    std::string importFile = "import.txt";

    std::vector<std::string> nonOptions;

};

Options parseOptions(int argc, char** argv, Verbosity verbosity = Verbosity::Warning) {
    Logger log(verbosity);
    Options result;
    static option long_options[] = {
        { "analyze",        no_argument,       nullptr, 'a' },
        { "export",         no_argument,       nullptr, 'e' },
        { "calculate",      no_argument,       nullptr, 'c' },

        { "repeat",         optional_argument, nullptr, 'r' },
        { "resource-theft", optional_argument, nullptr, 't' },
        { "stats",          required_argument, nullptr, 's' },
        { "import",         optional_argument, nullptr, 'i' },

        { "help",           no_argument,       nullptr, 'h' },
        { "verbose",        required_argument, nullptr, 'v' },
        { "version",        no_argument,       nullptr, 'V' },
        { nullptr,          0,                 nullptr, 0   }
    };
    int optionChar = 0;
    int optionIndex = 0;
    opterr = 0;
    while ((optionChar = getopt_long(argc, argv, ":acei::hr::t::s:v:V", long_options, &optionIndex)) != -1) {
        switch (optionChar) {
        case 'a': {
            if (result.calculate) {
                log.error("Cannot enable both --calculate and --analyze, ignoring --analyze\n");
            }
            else if (result.exportCompositions) {
                log.error("Cannot enable both --exportCompositions and --analyze, ignoring --analyze\n");
            }
            else {
                result.analyze = true;
            }
        } break;
        case 'c': {
            if (result.analyze) {
                log.error("Cannot enable both --analyze and --calculate, ignoring --calculate\n");
            }
            else if (result.exportCompositions) {
                log.error("Cannot enable both --exportCompositions and --calculate, ignoring --calculate\n");
            }
            else {
                result.calculate = true;
            }
        } break;
        case 'e': {
            if (result.analyze) {
                log.error("Cannot enable both --analyze and --exportCompositions, ignoring --exportCompositions\n");
            }
            else if (result.calculate) {
                log.error("Cannot enable both --calculate and --exportCompositions, ignoring --exportCompositions\n");
            }
            else {
                result.exportCompositions = true;
            }
        } break;
        case 'h': {
            result.help = true;
        } break;
        case 'i': {
            // getopt_long does not support --optional argument or --o argument,
            // only --optional=argument and -oargument. This contrasts with
            // --required argument and -r argument, which do work. I don't like
            // that inconsistency, so I manually fix it.
            if (optind < argc) {
                char* a = argv[optind];
                std::size_t n = std::strlen(a);
                if (!optarg && !(n == 2 && a[0] == '-') && !(n > 2 && a[0] == '-' && a[1] == '-')) {
                    // We didn't get an argument from getopt and next token does not
                    // look like an option, so we use it as an argument.
                    optarg = a;
                    ++optind;
                }
            }
            if (optarg) {
                struct stat buffer;
                if (stat(optarg, &buffer) == 0) {
                    log.debug("Selecting import file %s\n", optarg);
                    result.import = true;
                    result.importFile = optarg;
                }
                else {
                    log.error("Failed to select import file %s: %s\n",
                              optarg,
                              std::strerror(errno));
                }
            }
            else {
                log.debug("Selecting default import file %s\n",
                          result.importFile.c_str());
                result.import = true;
            }
        } break;
        case 'r': {
            result.repeatSameTargetIndex = true;
            if (optarg) {
                const std::string arg(optarg);
                if (arg == "0" || arg == "no" || arg == "false") {
                    result.repeatSameTargetIndex = false;
                }
            }
        } break;
        case 's': {
            struct stat buffer;
            if (stat(optarg, &buffer) == 0) {
                log.debug("Selecting stats file %s\n", optarg);
                result.statsFile = optarg;
            }
            else {
                log.error("Failed to select stats file %s: %s\n",
                          optarg,
                          std::strerror(errno));
            }
        } break;
        case 't': {
            result.enableResourceTheft = true;
            if (optarg) {
                const std::string arg(optarg);
                if (arg == "0" || arg == "no" || arg == "false") {
                    result.enableResourceTheft = false;
                }
            }
        } break;
        case 'v': {
            log.verbosity = Verbosity::FromString(optarg);
            result.verbosity = log.verbosity;
            log.trace("Setting verbosity to %s\n", optarg);
        } break;
        case 'V': {
            result.version = true;
        } break;
        case '?': {
            log.error("Ignoring invalid option character '%c'\n", optopt);
            result.help = true;
        } break;
        case ':': {
            log.error("Ignoring option '%s' because it lacks its mandatory argument\n", optopt);
            result.help = true;
        } break;
        default: {
            log.error("Error parsing option with character code '0x%x' ('%c')\n", optionChar, optionChar);
            result.help = true;
        }
        }
    }
    if (result.calculate) {
        while (optind < argc) {
            char* c = argv[optind];
            result.nonOptions.push_back(c);
            ++optind;
        }
    }
    else {
        while (optind < argc) {
            log.error("Ignoring non-option argument %s\n",
                      argv[optind]);
            ++optind;
        }
    }
    if (!result.calculate && !result.exportCompositions) {
        result.analyze = true;
    }
    return result;
}

void help(const char* executableName) {
    Logger log(Verbosity::Warning);
    log.warn("Usage: %s -a [option...]\n"
             "       %s -c [option...] <fleet...> vs <fleet...>\n"
             "       %s -e [option...]\n"
             "Analyze ship stats for balance.\n"
             "\n"
             "  -a, --analyze               Do a stats analysis. Mutually exclusive with\n"
             "                              --calculate and --export. Default: Enabled, unless\n"
             "                              --calculate or --export is given.\n"
             "  -c, --calculate             Run a single bcalc. Mutually exclusive with\n"
             "                              --analyze and --export. Default: Disabled.\n"
             "  -e, --export                Generate fleet compositions only. Mutually\n"
             "                              exclusive with --analyze and --calculate. Default:\n"
             "                              Disabled.\n"
             "\n"
             "  -h, --help                  Display this help message and exit.\n"
             "  -i, --import=<file>         Instead of calculating fleet compositions, read\n"
             "                              them from the given file, which must be formatted\n"
             "                              the same was as the fleets produced by --export.\n"
             "  -r, --repeat=[true|false]   Enable or disable target repetition. If enabled,\n"
             "                              this causes ships to repeat the same T1/T2/T3 if\n"
             "                              flakking has caused some ships in that class to\n"
             "                              stay alive after the first time. If no argument is\n"
             "                              given, enable. Default: Enabled.\n"
             "  -t, --resource-theft=[true|false]\n"
             "                              Enable or disable resource theft. If enabled in\n"
             "                              calculate mode, this program will assume there's\n"
             "                              always plenty of resources to steal. In analyze\n"
             "                              mode, this program will add a significant amount\n"
             "                              of resource stealers to every roiding fleet as\n"
             "                              well. If no argument is given, enable. Default:\n"
             "                              Disabled.\n"
             "  -s, --stats=<file>          Select the file that contains the ship stats to\n"
             "                              analyze or calculate with. Default: ./stats.txt\n"
             "  -v, --verbose=<level>       Set the verbosity. Options: None, Error, Warning,\n"
             "                              Info, Debug, Trace. Default: Warning\n"
             "  -V, --version               Display version information\n"
             "\n"
             "In analysis mode, no non-option arguments are accepted.\n"
             "\n"
             "In single bcalc mode, the battle must be specified on the command line. The\n"
             "syntax is as follows:\n"
             "\n"
             "    Fleet <score> <value> <ship...> vs Fleet <score> <value> <ship...>\n"
             "\n"
             "The scores and values are those of the originating planet, not those of the\n"
             "fleet. Multiple fleets may be given, each preceded by the keyword 'Fleet'. Each\n"
             "fleet may have a number of ships, consisting of a number and a ship name from\n"
             "the stats file. Example:\n"
             "\n"
             "    bcalc -c -s r67.txt Fleet 234000 123000 1000 Harpy 100 Pegasus vs Fleet \\\n"
             "        345000 234000 5000 Phantom 300 Asteroid Fleet 567000 456000 5000 Phantom\n"
             "\n"
             "Ship names are case-sensitive. Ship numbers, scores and values must be written\n"
             "out in full, there are no shortcuts like '3m or '500k'.\n"
             "\n"
             "In export mode, no non-option arguments are accepted. The outputted numbers\n"
             "are values, not ship numbers. Redirect the standard out stream if you want to\n"
             "write fleet compositions to file.\n",
             executableName, executableName);
}

void version() {
    Logger log(Verbosity::Warning);
    log.warn("mzCalc v%d.%d.%d\n"
             "Copyright (C) 2016  Josse van der Plaat\n"
             "License GPLv2: GNU GPL version v2 or later <https://www.gnu.org/licenses/gpl-2.0.html>.\n"
             "This is free software: you are free to change and redistribute it.\n"
             "There is NO WARRANTY, to the extent permitted by law.\n",
             majorVersion, minorVersion, patchVersion);
}

int main(int argc, char** argv) {
    Options options = parseOptions(argc, argv);

    Logger log(options.verbosity);
    if (options.help) {
        help(argv[0]);
    }
    else if (options.version) {
        version();
    }
    else {
        Stats stats = readStats(options.statsFile, options.verbosity);

        if (options.analyze) {
            log.debug("Using import ? %d, file %s\n",
                      options.import,
                      options.importFile.c_str());
            std::vector<Fleet> fleets = options.import
                ? readFleets(stats,
                             options.importFile,
                             options.verbosity)
                : fleetCompositions(stats,
                                    options.enableResourceTheft,
                                    options.verbosity);
            analyzeStats(stats,
                         fleets,
                         options.repeatSameTargetIndex,
                         options.enableResourceTheft,
                         options.verbosity);
        }
        else if (options.calculate) {
            log.verbosity = Verbosity::Trace;
            Battle battle;
            bool ok = fillBattle(battle, stats, options.nonOptions, options.verbosity);
            if (ok) {
                bcalc(battle,
                      options.repeatSameTargetIndex,
                      options.enableResourceTheft,
                      options.verbosity);
                log.warn("%s\n", battle.toString().c_str());
            }
        }
        else if (options.exportCompositions) {
            for (const Fleet& f : fleetCompositions(stats,
                                                    options.enableResourceTheft,
                                                    options.verbosity)) {
                Maybe<const BattleFormation> pod = f.firstPod();
                log.warn("%s\t%s:\n",
                         pod->shipType.race.toString().c_str(),
                         pod->shipType.clazz.toString().substr(0, 2).c_str());
                for (const BattleFormation& bf : f.battleFormations) {
                    log.warn("  %ld\t%s\n",
                             bf.atStartOfBattle * bf.shipType.cost,
                             bf.shipType.name.c_str());
                }
            }
        }
    }
    return 0;
}
