// This file is part of mzCalc.
//
// This program is free software; you can redistribute it and/or modify it under
// the terms of the GNU General Public License version 2, as published by the
// Free Software Foundation.
//
// This program is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
// details.
//
// You should have received a copy of the GNU General Public License along with
// mzCalc. If not, see <http://www.gnu.org/licenses/>.
//
// This work is Copyright (C) 2016 by Josse van der Plaat
//
// Individual portions may be copyright by individual contributors, and are
// included in this collective work with permission of the copyright owners.

#include "parser.h"

#include "logger.h"

#include <fstream>
#include <regex>

namespace {

void addBuiltIn(Stats& stats, Verbosity verbosity) {
    Logger log(verbosity);

    if (stats.create("Asteroid",
                     Class::Asteroid,
                     std::array<Class, MAX_TARGETS>{},
                     GunType::None,
                     Init(0),
                     Guns(0),
                     Armor(50),
                     ERes(0),
                     Cost(20000),
                     Eta(99),
                     Race::None)) {
        log.debug("Created Asteroid\n");
    }
    else {
        log.error("Failed to create Asteroid.\n");
    }

    if (stats.create("Structure",
                     Class::Structure,
                     std::array<Class, MAX_TARGETS>{},
                     GunType::None,
                     Init(0),
                     Guns(0),
                     Armor(500),
                     ERes(0),
                     Cost(20000),
                     Eta(98),
                     Race::None)) {
        log.debug("Created Structure\n");
    }
    else {
        log.error("Failed to create Structure.\n");
    }

    if (stats.create("Resource",
                     Class::Resource,
                     std::array<Class, MAX_TARGETS>{},
                     GunType::None,
                     Init(0),
                     Guns(0),
                     // The armor of 1 resource is 1/50, but that can't be
                     // encoded in an integer type. If this value is ever used,
                     // it will probably throw a divide-by-zero exception.
                     // Resource capture uses a hardcoded armor value instead.
                     Armor(0),
                     ERes(0),
                     Cost(1),
                     Eta(97),
                     Race::None)) {
        log.debug("Created Resource\n");
    }
    else {
        log.error("Failed to create Resource.\n");
    }
}

} // namespace

Stats readStats(const std::string& statsFile, Verbosity verbosity) {
    Logger log(verbosity);
    Stats stats;

    Race raceFromHeader;
    std::ifstream in(statsFile);
    std::string line;
    std::regex emptyLineRegex("[\t ]*", std::regex::optimize);
    std::regex raceRegex("(Terran|Cathaar|Xandathrii|Zikonian|Eitraides)", std::regex::optimize);
    std::regex headerRegex("Name\tClass\t.*", std::regex::optimize);
    std::regex shipRegex("([A-Za-z ]+)\t\
(Fighter|Corvette|Frigate|Destroyer|Cruiser|Battleship)\t\
((?:Fi|Co|Fr|De|Cr|Bs|Ro|St|Re|-|\t)*)\
(Norm|Emp|EMP|Cloak|Steal|Pod|Struc|Res)\t\
([0-9]+)\t\
([0-9]+)\t\
([0-9]+)\t\
([0-9]+|-)\t\
([0-9]+)\t\
([0-9]+)\t\
([0-9]+)\t\
([0-9]+)\t\
([0-9]+)\t\
([0-9]+)\t\
([0-9]+)?\t?\
(Ter|Cat|Xan|Zik|Etd)?\r?", std::regex::optimize);
    bool ok = true;
    bool done = false;
    while (ok && !done) {
        line.resize(127, '\0');
        in.getline(&line[0], line.size());
        ok = !in.fail();
        done = in.eof();
        if (ok) {
            line.resize(in.gcount() - 1);
            log.trace("Read complete line: '%s'\n",
                      line.c_str());
            std::smatch results;
            if (std::regex_match(line, results, emptyLineRegex)) {
                log.trace("Empty line\n");
            }
            else if (std::regex_match(line, results, raceRegex)) {
                raceFromHeader = Race(results[1]);
                log.trace("Race header:%s\n",
                          raceFromHeader.toString().c_str());
            }
            else if (std::regex_match(line, results, headerRegex)) {
                log.trace("Column header\n");
            }
            else if (std::regex_match(line, results, shipRegex)) {
                // 1:Name	2:Class	3:T1/2/3	4:Type	5:Init	6:Gun	7:Arm	8:Dmg	9:E/R	10:M	11:C	12:E	13:A/C	14:D/C	15:ETA?	16:Race?
                std::string name(             results[1]);
                Class       clazz(            results[2]);
                std::istringstream targetsStream(results[3]);
                std::vector<std::string> targetStrings(std::istream_iterator<std::string>{targetsStream},
                                                       std::istream_iterator<std::string>{});
                std::array<Class, MAX_TARGETS> targets;
                for (std::size_t i = 0; i < std::min(targetStrings.size(), targets.size()); ++i) {
                    targets[i] = Class(targetStrings[i]);
                }
                for (std::size_t i = targets.size(); i < targetStrings.size(); ++i) {
                    log.error("Application does not support more than %ld target indices, ignoring superfluous target '%s' at index %ld\n",
                              MAX_TARGETS,
                              targetStrings[i].c_str(),
                              i);
                }
                GunType     type(             results[4]);
                Init        init(  std::stoll(results[5]));
                Guns        guns(results[8] == "-"
                                 ? std::stoll(results[6])
                                 : std::stoll(results[8] /* Damage */));
                Armor       armor( std::stoll(results[7]));
                ERes        eres(  std::stoll(results[9]));
                Cost        cost(  std::stoll(results[10]) + std::stoll(results[11]) + std::stoll(results[12]));
                // int64_t  ac(    std::stoll(results[13]));
                // int64_t  dc(    std::stoll(results[14]));
                Eta         eta(100);
                if (results[15].matched) {
                    eta = Eta(     std::stoll(results[15]));
                }
                else {
                    switch (clazz.clazz) {
                    case Class::Fighter: case Class::Corvette: {   eta = Eta(12); } break;
                    case Class::Frigate: case Class::Destroyer: {  eta = Eta(13); } break;
                    case Class::Cruiser: case Class::Battleship: { eta = Eta(14); } break;
                    case Class::None: case Class::Asteroid: case Class::Structure: case Class::Resource: { /* Ignore. */ } break;
                    }
                }
                Race        race;
                if (results[16].matched) {
                    race = Race(              results[16]);
                }
                else {
                    race = raceFromHeader;
                }
                log.trace("Name:%s, Class:%s, T:%s/%s/%s, GunType:%s, Init:%ld, Guns:%ld, Armor:%ld, ERes:%ld, Cost:%ld, ETA:%ld, Race:%s\n",
                          name.c_str(),
                          clazz.toString().c_str(),
                          targets[0].toString().c_str(),
                          targets[1].toString().c_str(),
                          targets[2].toString().c_str(),
                          type.toString().c_str(),
                          init,
                          guns,
                          armor,
                          eres,
                          cost,
                          eta,
                          race.toString().c_str());

                if (stats.create(name, clazz, targets, type, init, guns, armor, eres, cost, eta, race)) {
                    log.debug("Created %s\n",
                              name.c_str());
                }
                else {
                    log.error("Failed to create %s.\n",
                              name.c_str());
                }
            }
            else {
                log.error("Failed to parse line '%s'!\n",
                          line.c_str());
            }
        }
        else if (!done) {
            log.error("Failed to read complete line! Got: '%s'\n",
                      line.c_str());
            log.trace("State: good:%d eof:%d fail:%d bad:%d\n",
                      in.good(),
                      in.eof(),
                      in.fail(),
                      in.bad());
        }
    }
    log.trace("End state: good:%d eof:%d fail:%d bad:%d\n",
              in.good(),
              in.eof(),
              in.fail(),
              in.bad());

    addBuiltIn(stats, verbosity);

    return stats;
}

std::vector<Fleet> readFleets(const Stats& stats, const std::string& file, Verbosity verbosity) {
    std::vector<Fleet> fleets;
    Logger log(verbosity);

    std::ifstream in(file);
    std::string line(240, '\0');
    std::regex battleFormationRegex("\\s+([0-9]+)\\s+([A-Za-z ]+)\r?",
                                    std::regex::optimize);
    std::regex newFleetRegex("(([A-Z][a-z]{2})\\s+[A-Za-z/+]+):\r?",
                             std::regex::optimize);
    bool ok = true;
    bool done = false;
    Fleet* lastFleet = nullptr;
    while (ok && !done) {
        line.resize(240, '\0');
        in.getline(&line[0], line.size());
        ok = !in.fail();
        done = in.eof();
        if (ok) {
            line.resize(in.gcount() - 1);
            log.trace("Read complete line: '%s'\n", line.c_str());
            // 1:Value	2:Name
            std::smatch results;
            if (std::regex_match(line, results, battleFormationRegex)) {
                if (lastFleet) {
                    Cost cost(std::stoll(results[1]));
                    std::string name(results[2]);
                    log.trace("Cost:%ld Name:%s\n",
                              cost,
                              name.c_str());
                    if (Maybe<const ShipType> ship = stats[name]) {
                        NumberOfShips num = cost / ship->cost;
                        lastFleet->addBattleFormation(*ship, num);
                        log.debug("Added %ld %s to fleet (cost %ld).\n",
                                  num,
                                  name.c_str(),
                                  cost);
                    }
                    else {
                        log.error("Failed to add %s to fleet: no such ship exists!\n",
                                  name.c_str());
                        ok = false;
                    }
                }
                else {
                    log.error("Failed to add '%s' to fleet: no fleet has been created yet!\n",
                              line.c_str());
                }
            }
            else if (std::regex_match(line, results, newFleetRegex)) {
                log.trace("Creating new fleet for %s...",
                          results[1].str().c_str());
                // Finish the old fleet.
                if (lastFleet) {
                    lastFleet->value = lastFleet->fleetCost() / 90;
                    lastFleet->xp    = lastFleet->value / 120;
                    lastFleet->score = lastFleet->value + 60 * lastFleet->xp;
                }
                fleets.emplace_back(results[1].str(), 0, Side::Defense);
                lastFleet = &fleets.back();
                lastFleet->race = Race::fromString(results[2]);
                log.trace(" Done.\n");
            }
            else {
                log.error("Failed to parse line '%s'\n",
                          line.c_str());
            }
        }
        else if (!done) {
            log.error("Failed to read complete line! Got: '%s'\n",
                      line.c_str());
            log.trace("State: good:%d eof:%d fail:%d bad:%d\n",
                      in.good(),
                      in.eof(),
                      in.fail(),
                      in.bad());
        }
    }
    // Finish the last fleet.
    if (lastFleet) {
        lastFleet->value = lastFleet->fleetCost() / 90;
        lastFleet->xp = lastFleet->value / 120;
        lastFleet->score = lastFleet->value + 60 * lastFleet->xp;
    }

    log.trace("End state: good:%d eof:%d fail:%d bad:%d\n",
              in.good(),
              in.eof(),
              in.fail(),
              in.bad());

    return fleets;
}
